/*
 * Grundkurs Informatik - Beispielprogramm zur LZW-Datenkompression
 * J. Schmidt, 2016
 *
 * Anmerkungen:
 * - das zu codierende Wort und die initialen Zeichen der Tabelle sind hier fest codiert,
 *   d.h. sie können nicht zur Laufzeit eingegeben werden.
 *   Anzupassen sind die Variablen "InitZeichen" und "eingabe"
 *
 * - es wurden hier absichtlich keine der eingebauten Stringfunktionen verwendet,
 *   damit die Kompression beliebiger Daten möglich ist (man hat sonst Probleme mit dem Nullbyte).
 *   Hierfür wurden separate Funktionen (Bytekette) implementiert, die im Grunde einem Datentyp String
 *   mit zur Laufzeit variabler Länge entsprechen
*/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "bytekette.h"
#include "lzw.h"


const unsigned int ANZ_BIT_CODE = 4;  // Länge eines Codeworts in Bit; maximal 16
const bool LOESCHECODETAB = false;    // true: lösche Codetabelle wenn voll

t_byte InitZeichen[] = {'A', 'H', 'K', 'M', 'N', 'P', 'U'};  // Zeichen, die in die initiale Codetabelle aufgenommen werden sollen
//t_byte InitZeichen[] = {'1', '2', '3', '4', '5', '6'};  // Zeichen, die in die initiale Codetabelle aufgenommen werden sollen


void initAusgabe(t_Ausgabe *a, const unsigned int anzBitCode)
{
	a->speicherBlockGroesse = 100;
	a->anzBitCode = anzBitCode;
	a->anzCodewoerter = 0;
	a->maxAnzCodewoerter = 0;
	a->ausgabe_kompakt = NULL;
	a->ausgabe_ui = NULL;
}

void speicherAnlegenAusgabe(t_Ausgabe *ausgabe)
{
	ausgabe->ausgabe_ui = (unsigned int *) malloc(ausgabe->speicherBlockGroesse * sizeof(unsigned int));
	ausgabe->ausgabe_kompakt = (unsigned int *) malloc(ausgabe->speicherBlockGroesse * sizeof(unsigned int));

	assert(ausgabe->ausgabe_ui != NULL);
	assert(ausgabe->ausgabe_kompakt != NULL);

	ausgabe->maxAnzCodewoerter = ausgabe->speicherBlockGroesse;
}

void speicherVergroessernAusgabe(t_Ausgabe *ausgabe)
{
	unsigned int *p = (unsigned int *)realloc(ausgabe->ausgabe_ui, (ausgabe->maxAnzCodewoerter + ausgabe->speicherBlockGroesse) * sizeof(unsigned int));
	assert(p != NULL);

	ausgabe->ausgabe_ui = p;

	p = (unsigned int *)realloc(ausgabe->ausgabe_kompakt, (ausgabe->maxAnzCodewoerter + ausgabe->speicherBlockGroesse) * sizeof(unsigned int));
	assert(p != NULL);

	ausgabe->ausgabe_kompakt = p;
	ausgabe->maxAnzCodewoerter = ausgabe->maxAnzCodewoerter + ausgabe->speicherBlockGroesse;
}

// generiere initiale Codetabelle
t_Codetabelle genCodeTabelle(const unsigned int anzBitCode, const t_byte zeichen[], const unsigned int anzZeichen)
{
	t_Codetabelle Codetabelle;
	t_eintrag_codetabelle *pTabelle = NULL;
	unsigned int anzTabellenEintraege = (unsigned int) pow(2.0, (double)anzBitCode);
	unsigned int i;

	assert(anzTabellenEintraege >= anzZeichen);
	assert(anzBitCode <= 16);

	pTabelle = (t_eintrag_codetabelle *) malloc(anzTabellenEintraege * sizeof(t_eintrag_codetabelle));
	assert(pTabelle != NULL);

	// Initalisierung mit Einzelzeichen
	for(i = 0; i < anzZeichen; i++)
	{
		initByteKette(&(pTabelle[i].zeichen));
		speicherAnlegenByteKette(&(pTabelle[i].zeichen));
		einzelZeichenAnhaengenByteKette(&(pTabelle[i].zeichen), zeichen[i]);
		pTabelle[i].code = i;
	}

	// Rest initialisieren
	for(i = anzZeichen; i < anzTabellenEintraege; i++)
	{
		initByteKette(&(pTabelle[i].zeichen));
		speicherAnlegenByteKette(&(pTabelle[i].zeichen));
	}

	Codetabelle.anzEintraege = anzZeichen;
	Codetabelle.tabelle = pTabelle;
	Codetabelle.anzBitCode = ANZ_BIT_CODE;
	Codetabelle.maxAnzEintraege = anzTabellenEintraege;

	return Codetabelle;
}

void zeigeCodeTabelle(const t_Codetabelle Codetabelle)
{
	unsigned int i;

	for(i = 0; i < Codetabelle.anzEintraege; i++)
	{
		printfByteKetteRechts(Codetabelle.tabelle[i].zeichen, 10);
		printf("    ");
		printf("%u\n", Codetabelle.tabelle[i].code);
	}
}

void zeigeAusgabe(const t_Ausgabe ausgabe)
{
	unsigned int i;

	for(i = 0; i < ausgabe.anzCodewoerter; i++)
		printf("%u, ", ausgabe.ausgabe_ui[i]);

	printf("\n");
	printf("Speicherplatzbedarf: %lu Bit\n", ausgabe.anzCodewoerter * ausgabe.anzBitCode);
}

void codeTabelleFreigeben(t_Codetabelle Codetabelle)
{	
	unsigned int i;

	for(i = 0; i < Codetabelle.maxAnzEintraege; i++)
	{
		speicherFreigebenByteKette(&(Codetabelle.tabelle[i].zeichen));
	}

	free(Codetabelle.tabelle);
}

void speicherFreigeben(t_Codetabelle Codetabelle, t_Ausgabe ausgabe)
{
	codeTabelleFreigeben(Codetabelle);

	assert(ausgabe.ausgabe_ui != NULL);
	free(ausgabe.ausgabe_ui);
	assert(ausgabe.ausgabe_kompakt != NULL);
	free(ausgabe.ausgabe_kompakt);
}

// Prüfe, ob k in Codetabelle enthalten ist
bool zeichenketteInCodetabelle(const t_Codetabelle Codetabelle, const t_byte_kette k)
{
	bool enthalten = false;
	unsigned int i = 0;

	while(enthalten == false && i < Codetabelle.anzEintraege)
	{
		if(byteKettenIdentisch(k, Codetabelle.tabelle[i].zeichen) == true)
			enthalten = true;
		i++;
	}

	return enthalten;
}

// Trage die Zeichenkette k in Codetabelle ein, wenn diese nicht voll ist
// liefert true, wenn erfolgreich eingetragen
bool codeEintragen(t_Codetabelle *Codetabelle, const t_byte_kette k, bool tabLoeschen)
{
	bool ok = false;
	unsigned int i = Codetabelle->anzEintraege;

	if(i < Codetabelle->maxAnzEintraege)
	{
		ersetzenByteKette(&(Codetabelle->tabelle[i].zeichen), k);
		Codetabelle->tabelle[i].code = i;
		Codetabelle->anzEintraege++;
		ok = true;
	}
	else if(tabLoeschen == true)
	{
		printf("\nAktuelle Codetabelle:\n");
		zeigeCodeTabelle(*Codetabelle);

		// neu initialisieren
		codeTabelleFreigeben(*Codetabelle);
		*Codetabelle = genCodeTabelle(ANZ_BIT_CODE, InitZeichen, sizeof(InitZeichen));
		printf("\nCodetabelle gelöscht\n");
	}

	return ok;
}

// Finde den Code für Zeichenkette k in der Codetabelle
// Rückgabe des Codes in "code"
// "code" ist gültig, wenn Rückgabewert == true
bool findeCode(const t_Codetabelle Codetabelle, const t_byte_kette k, unsigned int *code)
{
	unsigned int i = 0;
	bool gefunden = false;

	while(!gefunden && i < Codetabelle.anzEintraege)
	{
		if(byteKettenIdentisch(Codetabelle.tabelle[i].zeichen, k) == true)
		{
			*code = Codetabelle.tabelle[i].code;
			gefunden = true;
		}

		i++;
	}

	return gefunden;
}

// Gib den Code für Zeichenkette k aus
unsigned int codeAusgeben(t_Ausgabe *ausgabe, const t_Codetabelle Codetabelle, const t_byte_kette k)
{
	unsigned int code;
	bool gefunden;

	// suche Code in Tabelle
	gefunden = findeCode(Codetabelle, k, &code);
	assert(gefunden == true);

	// reicht der Speicher?
	if(ausgabe->anzCodewoerter == ausgabe->maxAnzCodewoerter)
		speicherVergroessernAusgabe(ausgabe);

	// Code ausgeben
	// kompakte Ausgabe fehlt noch
	ausgabe->ausgabe_ui[ausgabe->anzCodewoerter] = code;
	ausgabe->anzCodewoerter++;

	return code;
}

// Codiere die Bytes in "Eingabe" ("anzEingabeZeichen" Stück) mit LZW
// ausgabe enthält:
// ausgabe_ui: Array, das anschließend jeweils ein unsigned int mit einem Ausgabecode enthält (zur einfacheren Darstellung)
// ausgabe_kompakt: Array, das anschließend den Ausgabecode enthält, mit sowenig Speicher wie möglich (also ANZ_BIT_CODE Bit pro Codewort) - derzeit ungenutzt
// tabLoeschen: wenn true -> Lösche Codetabelle, wenn voll,  initialisiere neu
t_Ausgabe codiereLZW(const t_byte eingabe[], const unsigned int anzEingabeZeichen, t_Codetabelle *Codetabelle, bool tabLoeschen)
{
	t_Ausgabe ausgabe;  // Ausgabecode
	t_byte_kette praefix;
	t_byte_kette altes_praefix, eintragcodetab;
	t_byte aktZeichen;  // aktuelles Eingabezeichen

	unsigned int i;
	unsigned int ausgabeCode;

	// Speicher für Ausgabe reservieren
	initAusgabe(&ausgabe, Codetabelle->anzBitCode);
	speicherAnlegenAusgabe(&ausgabe);

	// Speicher für Präfix reservieren
	initByteKette(&praefix);
	speicherAnlegenByteKette(&praefix);
	initByteKette(&altes_praefix);
	speicherAnlegenByteKette(&altes_praefix);
	initByteKette(&eintragcodetab);
	speicherAnlegenByteKette(&eintragcodetab);

	printf("\nSchritt  Zeichen c  Präfix P  Eintrag in Tabelle   Ausgabe\n");

	for(i = 0; i < anzEingabeZeichen; i++)
	{
		// Lies nächstes Eingabezeichen c aus dem Eingabestring
		aktZeichen = eingabe[i];

		// aktuelles Präfix merken
		ersetzenByteKette(&altes_praefix, praefix);

		// Wenn Pc nicht in der Codetabelle gefunden wird
		einzelZeichenAnhaengenByteKette(&praefix, aktZeichen);
		if(zeichenketteInCodetabelle(*Codetabelle, praefix) == false)
		{
			//Gib den Code für P aus
			ausgabeCode = codeAusgeben(&ausgabe, *Codetabelle, altes_praefix);

			// Wenn Codetabelle noch nicht voll:
			// Trage Pc in die nächste freie Position ein
			codeEintragen(Codetabelle, praefix, tabLoeschen);

			// aktuelles Präfix merken  (nur für Ausgabe)
			ersetzenByteKette(&eintragcodetab, praefix);
			// setze P = c
			ersetzenByteKetteEinzel(&praefix, aktZeichen);

			printf("%4u        %c    ", i+1, aktZeichen);
			printfByteKetteRechts(praefix, 10);
			printf("  ");
			printfByteKetteRechts(eintragcodetab, 10);
			printf("      ");
			printfByteKetteRechts(altes_praefix, 10);
			printf(" =%u \n", ausgabeCode);
		}
		else
		{
			// der else-Zweig fehlt: Wenn Pc in Codetabelle enthalten,
			// dann setze P := Pc
			// dies wurde aber in der Zeile vor dem IF bereits gemacht
			printf("%4u        %c    ", i+1, aktZeichen);
			printfByteKetteRechts(praefix, 10);
			printf("           -                  -\n");
		}
	}

	//Gib den Code für P aus
	ausgabeCode = codeAusgeben(&ausgabe, *Codetabelle, praefix);
	printf("%4u        -             -           -      ", i+1);
	printfByteKetteRechts(praefix, 10);
	printf(" =%u \n", ausgabeCode);

	return ausgabe;
}


int main(int argc, char* argv[])
{
	// Eingabezeichenkette
	const char eingabe[] = "HUMUHUMUNUKUNUKUAPUAA";
//	const char eingabe[] = "11222211112222111114411133344333115555111155551111555511166116611661166116611661";
	unsigned int anzEingabeZeichen = strlen(eingabe);

	// Speicher für Ausgabecode
	t_Ausgabe ausgabe;

	// Codetabelle initialisieren
	t_Codetabelle Codetabelle = genCodeTabelle(ANZ_BIT_CODE, InitZeichen, sizeof(InitZeichen));

	printf("Initiale Codetabelle:\n");
	zeigeCodeTabelle(Codetabelle);

	ausgabe = codiereLZW((const t_byte*)eingabe, anzEingabeZeichen, &Codetabelle, LOESCHECODETAB);

	printf("Ausgabe:\n");
	zeigeAusgabe(ausgabe);

	printf("Finale Codetabelle:\n");
	zeigeCodeTabelle(Codetabelle);

	// aufraumen
	speicherFreigeben(Codetabelle, ausgabe);

	return 0;
}

